/*
 *
 *    This file is part of NSA ( NullSearchAccess ).
 *
 *  NSA ( NullSearchAccess ) is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  NSA ( NullSearchAccess ) is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with NSA ( NullSearchAccess ).  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <wait.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "global.h"
#include "config.h"

int
main (int argc, char **argv)
{
  ppid = getpid();
  debug = 1;
  proc_alive = 0;
  pile = NULL;
  program_name = argv[0];

  print_version();
  if (argc == 1)
    xexit(ERR_NO_ARGV);
  decode_getopt(argc, argv);
  
  return EXIT_SUCCESS;
}

void
decode_getopt (const int argc, char **argv)
{
  int i;
  int debug_on;
  int ok;
  char c;
  struct opt_s opt;
  struct proto_s proto[] =
    {
      {"ftp", "anonymous", "anonymous", 21},
      {"pop", "posmaster", "123456", 110},
      {"mysql", "root", "", 3306},
      {"imap", "root", "123456", 143},
      {NULL, NULL, NULL, 0}
    };

  memset(&opt, 0, sizeof(struct opt_s));
  p_opt = &opt;
  p_opt->n_proc = DEFAULT_PROC_NUMBERS;
  p_opt->timeout = DEFAULT_TIMEOUT;
  p_opt->port = -1;
  debug_on = 0;

  stt.start = 0;
  stt.stop = 0;
  stt.n_ip = 0;
  stt.n_open = 0;
  stt.n_found = 0;

  do
    {
      c = getopt(argc, argv, STR_OPT_LIST);
      switch (c)
        {
          case 'h':
            print_help();
            return;
            break;
          case 'd':
            debug_on = 1;
            break;
          case 'v':
            p_opt->verbose = 1;
            break;
          case 'l':
            strncpy(p_opt->login, optarg, SIZE_AUTHEN-1);
            break;
          case 'a':
            strncpy(p_opt->passwd, optarg, SIZE_AUTHEN-1);
            break;
          case 'p':
            p_opt->port = xstrtol(optarg, &ok);
            if (ok)
              xexit(ERR_PORT);
           break;
          case 't':
            p_opt->timeout = xstrtol(optarg, &ok);
            if (ok)
              xexit(ERR_STRTOL);
            break;
          case 'n':
            p_opt->n_proc = xstrtol(optarg, &ok);
            if (ok)
              xexit(ERR_STRTOL);
            break;
          case 'f':
            p_opt->log = 1;
            break;
          case 'b':
            strncpy(p_opt->ip_current, optarg, IPv4_SIZE-1);
            break;
          case 'e':
            strncpy(p_opt->ip_end, optarg, IPv4_SIZE-1);
            break;
          case 'm':
            strncpy(p_opt->proto, optarg, LEN_PROTO-1);
            break;
        }
    }
  while (c != -1);

  if (!p_opt->ip_current[0] || !p_opt->ip_end[0])
    xexit(ERR_IP_UNKNOW);

  ok = check_rang_ip();
  if (ok == -1)
    xexit(ERR_IP_INVALID);

  i = 0;
  ok = 0;
  while (proto[i].port)
    {
      if (!strcmp(p_opt->proto, proto[i].name))
        {
          if (!p_opt->login[0])
            strncpy(p_opt->login, proto[i].login, SIZE_AUTHEN-1);
          if (!p_opt->passwd[0])
            strncpy(p_opt->passwd, proto[i].passwd, SIZE_AUTHEN-1);
          if (p_opt->port == -1)
            p_opt->port = proto[i].port;
          ok++;
        }
      i++;
    }

  if (!ok)
    xexit(ERR_PROTO_UNKNOW);

  if (p_opt->verbose)
    print_verbose_argv();

  if (!debug_on)
    debug = 0;

  scanner();
  p_opt = NULL;
}

char *
xstrdup (const char *original)
{
  char *copy = NULL;

  copy = strdup(original);
  if (!copy)
    xexit(ERR_STRDUP);

  return copy;
}

void *
xmalloc (const int size)
{
  void *ptr = NULL;

  if (!size)
    return NULL;

  ptr = malloc(size);
  if (!ptr)
    xexit(ERR_ALLOC_MEM);

  return ptr;
}

int
xstrtol (const char *str, int *error)
{
  char *check = NULL;
  char *copy = NULL;
  int nb;

  if (error)
    *error = 0;

  copy = xstrdup(str);
  nb = strtol(copy, &check, 10);
  if (*check != '\0')
    {
      fprintf(stderr, "[strtol()] argument <%s> invalid.\n", str);
      if (error)
        *error = 1;
    }

  xfree(copy);
  return nb;
}

void
xfree (void *ptr)
{
  if (ptr)
    {
      free(ptr);
      ptr = NULL;
    }
}

int
check_rang_ip (void)
{
  int tab1[4] = {0};
  int tab2[4] = {0};
  int ok[2];
  int i;

  ok[0] = parse_tab_ip(p_opt->ip_current, tab1);
  ok[1] = parse_tab_ip(p_opt->ip_end, tab2);

  if (ok[0] == -1 || ok[1] == -1)
    return -1;

  i = 0;
  while (i < 4)
    {
      if (tab1[i] < 0 || tab1[i] > 255 || tab2[i] < 0 ||
           tab2[i] > 255 || tab1[i] > tab2[i])
        return -1;
      if (tab1[i] < tab2[i])
        return 0;
      i++;
    }

  return 0;
}

int
parse_tab_ip (const char *ip, int *tab)
{
  int i;
  int ok;
  char *cut = NULL;
  char copy[IPv4_SIZE] = {0};

  i = 0;
  strncpy(copy, ip, IPv4_SIZE-1);
  cut = strtok(copy, ".");
  while (cut && i < 4)
    {
      tab[i++] = xstrtol(cut, &ok);
      if (ok)
        return -1;
      cut = strtok(NULL, ".");
    }

  return (i < 4) ? -1 : 0;
}

void
scanner (void)
{
  key_t key;
  int id, ok;

  key = create_key();
  id = create_shared_memory(key, p_opt->n_proc*IPv4_SIZE);
  ip_scanned = p_opt->ip_current;
  signal_init();

  ok = 1;
  print_started();
  stt.start = time(NULL);

  while (ok)
    {
      if (proc_alive == p_opt->n_proc)
        RIP_child(id, 1);
      else
        RIP_child(id, 0);
      ok = strcmp(p_opt->ip_current, p_opt->ip_end);
      start_child(id);
    }

  while (proc_alive)
    RIP_child(id, 0);

  stt.stop = time(NULL);
  shared_memory_destroy(id);
  print_static();
  free_pile();
}

void
start_child (int id)
{
  pid_t pid;
  int ret;
  char *memory = NULL;

  pid = fork();
  if (!pid)
    {
      signal(SIGQUIT, SIG_IGN);
      switch (p_opt->proto[0])
        {
          case 'f':
            ret = test_ftp_access();
            break;
          case 'p':
            ret = test_pop3_access();
            break;
          case 'm':
            ret = test_mysql_access();
            break;
          case 'i':
            ret = test_imap_access();
            break;
        }
      if (!ret)
        {
          memory = attach_shared_memory(id);
          if (!memory)
            fprintf(stderr, "mode verbose forced : %s accept null session\n", p_opt->ip_current);
          else
            {
              snprintf(memory, IPv4_SIZE+1, "%s ", p_opt->ip_current);
              if (p_opt->verbose)
                printf("mode verbose : %s accept null session\n", p_opt->ip_current);
              detach_shared_memory(memory);
            }
        }
      free_pile();
      exit(ret);
    }
  else if (pid != -1 && pid)
    {
      ip_scanned = p_opt->ip_current;
      incremente_ip(p_opt->ip_current);
      proc_alive++;
      stt.n_ip++;
    }
  else if (pid == -1 && debug)
    perror("[fork()] ");

}

void
RIP_child (int id, int forced)
{
  char *memory = NULL;
  char *p_mem = NULL;
  char ip[IPv4_SIZE] = {0};
  int ret;
  int i;
  int status;

  if (forced)
    wait(&status);
  else
    {
      ret = waitpid(-1, &status, WNOHANG);
      if (ret < 1)
        return;
    }

  proc_alive--;
  ret = WEXITSTATUS(status);
  if (!ret)
    {
      memory = attach_shared_memory(id);
      if (!memory)
        return;

      i = 0;
      p_mem = memory;
      while (*p_mem)
        {
          ip[i++] = *p_mem++;
          if (*p_mem == ' ')
            {
              ip[i] = '\0';
              push_pile(ip);
              if (p_opt->log)
                write_file (ip);
              memset(ip, 0, strlen(ip));
              i = 0;
              p_mem++;
            }
        }
      stt.n_found++;
      memset(memory, 0, strlen(memory));
      detach_shared_memory(memory);
    }
  if (ret == 2 || !ret)
    stt.n_open++;
}

int
incremente_ip (char *ip)
{
  char *cut = NULL;
  int tab[4] = {0};
  int i;

  i = 0;
  cut = strtok(ip, ".");
  while (cut)
    {
      tab[i] = atoi(cut);
      cut = strtok(NULL, ".");
      i++;
    }

  tab[3]++;
  if (tab[3] > 255)
    {
      tab[3] = 0;
      tab[2]++;
    }
  if (tab[2] > 255)
    {
      tab[2] = 0;
      tab[1]++;
    }
  if (tab[1] > 255)
    {
      tab[1] = 0;
      tab[0]++;
    }
  if (tab[0] > 255)
    return -1;

  snprintf(ip, IPv4_SIZE, "%d.%d.%d.%d",
              tab[0], tab[1], tab[2], tab[3]);
  return 0;
}

void
push_pile (const char *ip)
{
  struct pile_s *new = NULL;
  struct pile_s *p_pile = pile;

  new = xmalloc(sizeof(struct pile_s));
  strncpy(new->ip, ip, IPv4_SIZE);
  new->next = NULL;

  if (!pile)
    pile = new;
  else
    {
      while (p_pile->next)
        p_pile = p_pile->next;
      p_pile->next = new;
    }
}

void
free_pile (void)
{
  struct pile_s *buff = NULL;

  while (pile)
    {
      buff = pile->next;
      xfree(pile);
      pile = buff;
    }
}

void
signal_init (void)
{
  struct sigaction act;
  int i;
  int siglist[] =
    {
      SIGUSR1,
      SIGUSR2,
      SIGINT,
      SIGPIPE,
      SIGQUIT,
      SIGTERM,
      SIGSTOP,
      SIGTSTP,
      SIGKILL,
      SIGHUP,
      SIGILL,
      SIGABRT,
      SIGFPE,
      SIGSEGV,
      SIGALRM,
      SIGCHLD,
      SIGCONT,
      SIGTTIN,
      SIGTTOU,
      0
    };

  i = 0;
  memset(&act, 0, sizeof(struct sigaction));
  act.sa_handler = signal_handler;

  while (siglist[i])
    sigaction(siglist[i++], &act, NULL);
}

void
signal_handler (int status)
{
  if (status == SIGINT && ppid == getpid())
    {
      printf("Signal SIGINT receved ... wait finish child for stop program\n");
      while (proc_alive)
        {
          wait(NULL);
          proc_alive--;
        }
      ip_scanned = NULL;
      exit(EXIT_FAILURE);
    }
  if (status == SIGPIPE && ppid == getpid())
    fprintf(stderr, "recv signal SIGPIPE socket closed.\n");
  if (status == SIGQUIT && ppid == getpid())
    print_static_current();
}

int
check_answer_proto (const char *answer, const char *original_answer)
{
  int ret;
  const char *p_answer = NULL;
  const char *p_original = NULL;

  ret = 0;
  p_answer = answer;
  p_original = original_answer;
  while (*p_answer && *p_original && ret != -1)
    {
      if (*p_answer++ != *p_original++)
        ret = -1;
    }

  p_answer = NULL;
  p_original = NULL;
  return ret;
}
   
void
write_file (const char *ip)
{
  int fd;
  int ret;
  char str[300] = {0};

  fd = open(LOG_NAME, O_CREAT | O_APPEND | O_WRONLY, S_IRWXU);
  if (fd == -1 && debug)
    {
      perror("[open()] ");
      return;
    }

  snprintf(str, sizeof(str)-1, "IP/Login/Passwd: %s/%s/%s .\n", ip, p_opt->login, p_opt->passwd);
  ret = write(fd, str, strlen(str));
  if (ret == -1 && debug)
    perror("[write()] ");

  if (fd != -1)
    close(fd);
}
