/*
 *
 *    This file is part of NSA ( NullSearchAccess ).
 *
 *  NSA ( NullSearchAccess ) is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  NSA ( NullSearchAccess ) is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with NSA ( NullSearchAccess ).  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <errno.h>
#include <stdio.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>

#include "global.h"
#include "typedef.h"

struct socket_s *
create_socket (void)
{
  struct socket_s *sock = NULL;

  sock = xmalloc(sizeof(struct socket_s));
  sock->socket = create_socket_stream();
  if (sock->socket == -1)
    {
      xfree(sock);
      return NULL;
    }

  sock->to.sin_family = AF_INET;
  sock->to.sin_port = htons(p_opt->port);
  sock->to.sin_addr.s_addr = inet_addr(p_opt->ip_current);

  return sock;
}

int
create_socket_stream (void)
{
  int socket_stream;

  socket_stream = socket(AF_INET, SOCK_STREAM, 0);
  if (socket_stream == -1 && debug)
      perror("[socket()] ");

  return socket_stream;
}

int
connect_socket(struct socket_s *sock)
{
  struct timeval time;
  int ret;
  int res;
  fd_set fd_read;
  socklen_t size;

  time.tv_sec = p_opt->timeout;
  time.tv_usec = 0;

  FD_ZERO(&fd_read);
  FD_SET(sock->socket, &fd_read);

  ret = block_or_unblock_socket (sock->socket);
  if (ret == -1)
    return -1;

  ret = connect(sock->socket, (struct sockaddr*)&sock->to, sizeof(sock->to));
  if (ret == -1)
    {
      ret = select(sock->socket+1, &fd_read, NULL, NULL, &time);
      if (ret != 1)
        return -1;
    }

  size = sizeof(int);
  ret = getsockopt(sock->socket, SOL_SOCKET, SO_ERROR, &res, &size);
  if (res || ret == -1)
    {
      if (ret == -1 && debug)
        perror("[getsockopt()] ");
      return -1;
    }

  ret = block_or_unblock_socket (sock->socket);
  return ret;
}

int
xsend (const char *str, int size, int socket)
{
  int ret;

  ret = send(socket, str, size, 0);
  if (ret == -1 && debug)
    perror("[send()] ");
      
  return ret;
}

int
xrecv (char *str, size_t size, int socket)
{
  struct timeval time;
  fd_set fd_read;
  int ret;

  time.tv_sec = 7; 
  time.tv_usec = 0;

  FD_ZERO(&fd_read);
  FD_SET(socket, &fd_read);

  memset(str, 0, size);
  ret = select(socket+1, &fd_read, NULL, NULL, &time);
  if (ret)
    {
      ret = recv(socket, str, size, 0);
      if (ret == -1 && debug)
        perror("[recv()] ");
    }

  return ret;
}

int
block_or_unblock_socket (int socket)
{
  int flags;
  int ret;

  flags = get_socket_mode(socket);
  if (flags == -1)
    return -1;
 
  flags |= O_NONBLOCK;
  ret = fcntl(socket, F_SETFL, flags);
  if (ret == -1 && debug)
    perror("[fcntl()] ");

  return ret;
}

int
get_socket_mode (int socket)
{
  int flags;

  flags = fcntl(socket, F_GETFL, 0);
  if (flags == -1 && debug)
    perror("[fcntl()] ");

  return flags;
}

void
free_socket (struct socket_s *socket)
{
  if (socket)
    {
      if (socket->socket != -1)
        close(socket->socket);
      xfree(socket);
    }
}